##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

import warnings
import inspect
from itertools import groupby
from typing import Dict, List, Tuple, Type
from azure.quantum import Workspace

try:
    from qiskit.providers import ProviderV1 as Provider
    from qiskit.providers.exceptions import QiskitBackendNotFoundError
    from qiskit.providers import BackendV1 as Backend
    from qiskit.exceptions import QiskitError
except ImportError:
    raise ImportError(
        "Missing optional 'qiskit' dependencies. \
To install run: pip install azure-quantum[qiskit]"
    )

from azure.quantum.qiskit.backends.backend import AzureBackendBase
from azure.quantum.qiskit.job import AzureQuantumJob
from azure.quantum.qiskit.backends import *

# Target ID keyword for parameter-free solvers
PARAMETER_FREE = "parameterfree"

QISKIT_USER_AGENT = "azure-quantum-qiskit"


class AzureQuantumProvider(Provider):
    """
    Class for interfacing with the Azure Quantum service
    using Qiskit quantum circuits
    """
    def __init__(self, workspace: Workspace=None, **kwargs):
        """AzureQuantumService class

        :param workspace: Azure Quantum workspace. If missing it will create a new Workspace passing `kwargs` to the constructor. Defaults to None. 
        :type workspace: Workspace
        """
        if kwargs is not None and len(kwargs) > 0:
            from warnings import warn
            warn(f"""Consider passing \"workspace\" argument explicitly. 
                 The ability to initialize AzureQuantumProvider with arguments {', '.join(f'"{argName}"' for argName in kwargs)} is going to be deprecated in future versions.""", 
                 DeprecationWarning, 
                 stacklevel=2)

        if workspace is None:
            workspace = Workspace(**kwargs)

        workspace.append_user_agent(QISKIT_USER_AGENT)

        self._workspace = workspace
        self._backends = None

    def get_workspace(self) -> Workspace:
        """Return Azure Quantum Workspace"""

        return self._workspace

    def get_backend(self, name=None, **kwargs) -> AzureBackendBase:
        """Return a single backend matching the specified filtering.

        Args:
            name (str): name of the backend.
            **kwargs: dict used for filtering.
        Returns:
            Backend: a backend matching the filtering.
        Raises:
            QiskitBackendNotFoundError: if no backend could be found or
                more than one backend matches the filtering criteria.
        """

        backends = self.backends(name=name, **kwargs)

        if len(backends) > 1:
            def all_same(iterable):
                group_iter = groupby(iterable)
                return next(group_iter, True) and not next(group_iter, False)

            # If all backends have the same name, filter for default backend
            if all_same(backend.name() for backend in backends):
                backends = list(
                    filter(
                        lambda backend: self._match_all(
                            backend.configuration().to_dict(), {"is_default": True}
                        ),
                        backends,
                    )
                )

        if len(backends) > 1:
            raise QiskitBackendNotFoundError(
                "More than one backend matches the criteria"
            )
        if not backends:
            raise QiskitBackendNotFoundError(
                f"Could not find target '{name}'. \
Please make sure the target name is valid and that the associated provider is added to your Workspace. \
To add a provider to your quantum workspace on the Azure Portal, \
see https://aka.ms/AQ/Docs/AddProvider"
            )
        return backends[0]

    def backends(self, name=None, **kwargs):
        """Return a list of backends matching the specified filtering.
        
        Args:
            name (str): name of the backend.
            **kwargs: dict used for filtering.
        Returns:
            typing.List[qiskit.providers.BackendV1]: a list of Backends that match the filtering
                criteria.
        """

        # Lazy load backends
        if self._backends is None:
            self._backends = self._init_backends()

        if name:
            if name not in self._backends:
                raise QiskitBackendNotFoundError(
                    f"The '{name}' backend is not installed in your system."
                )

        provider_id = kwargs.get("provider_id", None)

        allowed_targets = self._get_allowed_targets_from_workspace(name, provider_id)

        workspace_allowed = lambda backend: self._is_available_in_ws(
            allowed_targets, backend
        )

        # flatten the available backends
        backend_list = [x for v in self._backends.values() for x in v]

        # filter by properties specified in the kwargs and filter function
        backends: List[Backend] = self._filter_backends(
            backend_list, filters=workspace_allowed, **kwargs
        )

        return backends

    def get_job(self, job_id) -> AzureQuantumJob:
        """Returns the Job instance associated with the given id."""
        azure_job = self._workspace.get_job(job_id)
        backend = self.get_backend(azure_job.details.target)
        return AzureQuantumJob(backend, azure_job)

    def _is_available_in_ws(
        self, allowed_targets: List[Tuple[str, str]], backend: Backend
    ):
        for name, provider in allowed_targets:
            if backend.name() == name:
                config = backend.configuration().to_dict()
                if "azure" in config and "provider_id" in config["azure"]:
                    if config["azure"]["provider_id"] == provider:
                        return True
        return False

    def _get_allowed_targets_from_workspace(
        self, name: str, provider_id: str
    ) -> List[Tuple[str, str]]:
        target_statuses = self._workspace._get_target_status(name, provider_id)
        candidates: List[Tuple[str, str]] = []
        for provider_id, status in target_statuses:
            candidates.append((status.id, provider_id))
        return candidates

    def _get_candidate_subclasses(self, subtype: Type[Backend]):
        if not inspect.isabstract(subtype):
            yield subtype

        subclasses = subtype.__subclasses__()
        if subclasses:
            for subclass in subclasses:
                for leaf in self._get_candidate_subclasses(subclass):
                    yield leaf


    def _init_backends(self) -> Dict[str, List[Backend]]:
        instances: Dict[str, List[Backend]] = {}
        subclasses = list(self._get_candidate_subclasses(subtype=AzureBackendBase))

        for backend_cls in subclasses:
            backend_names = self._backend_names(backend_cls)
            if backend_names is None:
                continue

            for name in backend_names:
                backend_instance: Backend = self._get_backend_instance(
                    backend_cls, name
                )
                backend_name: str = backend_instance.name()
                instances.setdefault(backend_name, []).append(backend_instance)

        return instances

    def _backend_names(self, backend_cls: Type[Backend]) -> List[str]:
        if hasattr(backend_cls, "backend_names"):
            return list(backend_cls.backend_names)
        return None

    def _get_backend_instance(self, backend_cls: Type[Backend], name: str) -> Backend:
        try:
            return backend_cls(name=name, provider=self)
        except Exception as err:
            raise QiskitError(
                f"Backend {backend_cls} could not be instantiated: {err}"
            ) from err

    def _match_all(self, obj, criteria):
        """Return True if all items in criteria matches items in obj."""
        return all(
            self._match_config(obj, key_, value_) for key_, value_ in criteria.items()
        )

    def _match_config(self, obj, key, value):
        """Return True if the criteria matches the base config or azure config."""
        return obj.get(key, None) == value or self._match_azure_config(
            obj, key, value
        )

    def _match_azure_config(self, obj, key, value):
        """Return True if the criteria matches the azure config."""
        azure_config = obj.get("azure", {})
        return azure_config.get(key, None) == value

    def _has_config_value(self, obj, key):
        """Return True if the key is found in the root config or azure config."""
        return key in obj or key in obj.get("azure", {})

    def _filter_backends(
        self, backends: List[Backend], filters=None, **kwargs
    ) -> List[Backend]:
        """Return the backends matching the specified filtering.
        Filter the `backends` list by their `configuration` attributes,
        or from a boolean callable. The criteria for filtering can
        be specified via `**kwargs` or as a callable via `filters`, and the
        backends must fulfill all specified conditions.
        
        Args:
            backends (list[Backend]): list of backends.
            filters (callable): filtering conditions as a callable.
            **kwargs: dict of criteria.
        Returns:
            list[Backend]: a list of backend instances matching the
                conditions.
        """

        configuration_filters = {}
        unknown_filters = {}
        for key, value in kwargs.items():
            # If `any` of the backends has the key in its configuration, filter by it.
            # qiskit API for this requires `all` backends to have the key in
            # their configuration to be considered for filtering
            print(f"Looking for {key} with {value}")
            if any(
                self._has_config_value(backend.configuration().to_dict(), key) for backend in backends
            ):
                configuration_filters[key] = value
            else:
                unknown_filters[key] = value

        if configuration_filters:
            backends = list(
                filter(
                    lambda backend: self._match_all(
                        backend.configuration().to_dict(), configuration_filters
                    ),
                    backends,
                )
            )

        if unknown_filters:
            warnings.warn(
                f"Specified filters {unknown_filters} are not supported by the available backends."
            )

        backends = list(filter(filters, backends))
        
        return backends



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

from urllib.request import urlopen

def df_chemistry() -> bytes:
    """
    Returns bitcode of a QIR program for the double-factorized chemistry
    quantum algorithm.
    """
    return urlopen("https://aka.ms/RE/df_chemistry").read()



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from enum import Enum
from azure.core import CaseInsensitiveEnumMeta


class DimensionScope(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The scope at which the quota is applied."""

    WORKSPACE = "Workspace"
    SUBSCRIPTION = "Subscription"


class ItemType(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The type of item."""

    JOB = "Job"
    """A program, problem, or application submitted for processing."""
    SESSION = "Session"
    """A logical grouping of jobs."""


class JobStatus(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The status of the job."""

    WAITING = "Waiting"
    EXECUTING = "Executing"
    SUCCEEDED = "Succeeded"
    FAILED = "Failed"
    CANCELLED = "Cancelled"


class JobType(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The type of job."""

    UNKNOWN = "Unknown"
    QUANTUM_COMPUTING = "QuantumComputing"
    OPTIMIZATION = "Optimization"


class JsonPatchOperation(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The operation to be performed."""

    ADD = "add"
    REMOVE = "remove"
    REPLACE = "replace"
    MOVE = "move"
    COPY = "copy"
    TEST = "test"


class MeterPeriod(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The time period in which the quota's underlying meter is accumulated. Based on calendar year.
    'None' is used for concurrent quotas.
    """

    NONE = "None"
    MONTHLY = "Monthly"


class ProviderAvailability(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """Provider availability."""

    AVAILABLE = "Available"
    DEGRADED = "Degraded"
    UNAVAILABLE = "Unavailable"


class SessionJobFailurePolicy(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """Policy controlling the behavior of the Session when a job in the session fails."""

    ABORT = "Abort"
    """New jobs submitted after a job fails will be rejected."""
    CONTINUE = "Continue"
    """New jobs submitted after a job fails will be accepted."""


class SessionStatus(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """The status of the session."""

    WAITING = "Waiting"
    EXECUTING = "Executing"
    SUCCEEDED = "Succeeded"
    FAILED = "Failed"
    FAILURE_S_ = "Failure(s)"
    TIMED_OUT = "TimedOut"


class TargetAvailability(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """Target availability."""

    AVAILABLE = "Available"
    DEGRADED = "Degraded"
    UNAVAILABLE = "Unavailable"



import warnings

from azure.quantum.job.base_job import ContentType
from azure.quantum.job.job import Job
from azure.quantum.target.target import Target
from azure.quantum.workspace import Workspace
from azure.quantum.target.params import InputParams
from typing import Any, Dict, Type, Union
from .job import MicrosoftElementsDftJob


class MicrosoftElementsDft(Target):
    """
    Microsoft Elements Dft target from the microsoft-elements provider.
    """

    target_names = [
        "microsoft.dft"
    ]


    def __init__(
        self,
        workspace: "Workspace",
        name: str = "microsoft.dft",
        **kwargs
    ):
        """
        Initializes a new DFT target.

        :param workspace: Associated workspace
        :type workspace: Workspace
        :param name: Target name
        """
        # There is only a single target name for this target
        assert name == self.target_names[0]

        # make sure to not pass argument twice
        kwargs.pop("provider_id", None)

        super().__init__(
            workspace=workspace,
            name=name,
            input_data_format="microsoft.xyz.v1",
            output_data_format="microsoft.dft-results.v1",
            provider_id="microsoft-elements",
            content_type=ContentType.text_plain,
            **kwargs
        )


    def submit(self,
               input_data: Any,
               name: str = "azure-quantum-dft-job",
               shots: int = None,
               input_params: Union[Dict[str, Any], InputParams, None] = None,
               **kwargs) -> MicrosoftElementsDftJob:
        """
        Submit DFT job to Azure Quantum Services.
        
        :param input_data: Input data
        :type input_data: Any
        :param name: Job name
        :type name: str
        :param shots: Number of shots. Ignored in DFT job. Defaults to None
        :type shots: int
        :param input_params: Input parameters
        :type input_params: Dict[str, Any]
        :return: Azure Quantum job
        :rtype: Job
        """

        if shots is not None:
            warnings.warn("The 'shots' parameter is ignored in Microsoft Elements Dft job.")
        
        return super().submit(
            input_data=input_data,
            name=name, 
            shots=shots, 
            input_params=input_params,
            **kwargs
        )


    @classmethod
    def _get_job_class(cls) -> Type[Job]:
        return MicrosoftElementsDftJob



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
from typing import TYPE_CHECKING, Any, Dict, Optional, Union, Type,  Protocol, runtime_checkable
import io
import json
import abc
import warnings

from azure.quantum._client.models import TargetStatus, SessionDetails
from azure.quantum._client.models._enums import SessionJobFailurePolicy
from azure.quantum.job.job import Job, BaseJob
from azure.quantum.job.session import Session, SessionHost
from azure.quantum.job.base_job import ContentType
from azure.quantum.target.params import InputParams
if TYPE_CHECKING:
    from azure.quantum import Workspace


@runtime_checkable
class QirRepresentable(Protocol):
    _name: str

    @abc.abstractmethod
    def _repr_qir_(self, **kwargs: Any) -> bytes:
        raise NotImplementedError


class Target(abc.ABC, SessionHost):

    _QSHARP_USER_AGENT = "azure-quantum-qsharp"

    """Azure Quantum Target."""
    # Target IDs that are compatible with this Target class.
    # This variable is used by TargetFactory. To set the default
    # target class for a given provider, specify the
    # default_targets constructor argument.
    #
    # If you provide a custom job class (derived from
    # azurem.quantum.job.job.Job) for this target, you must pass this type to
    # __init__ via the job_cls parameter.  This is then used by the target's
    # submit and get_job method.
    target_names = ()
    """Tuple of target names."""

    # Name of the provider's input parameter which specifies number of shots for a submitted job.
    # If None, target will not pass this input parameter. 
    _SHOTS_PARAM_NAME = None

    def __init__(
        self,
        workspace: "Workspace",
        name: str,
        input_data_format: str = "",
        output_data_format: str = "",
        capability: str = "",
        provider_id: str = "",
        content_type: ContentType = ContentType.json,
        encoding: str = "",
        average_queue_time: Union[float, None] = None,
        current_availability: str = ""
    ):
        """
        Initializes a new target.

        :param workspace: Associated workspace
        :type workspace: Workspace
        :param name: Target name
        :type name: str
        :param input_data_format: Format of input data (ex. "qir.v1")
        :type input_data_format: str
        :param output_data_format: Format of output data (ex. "microsoft.resource-estimates.v1")
        :type output_data_format: str
        :param capability: QIR capability
        :type capability: str
        :param provider_id: Id of provider (ex. "microsoft-qc")
        :type provider_id: str
        :param content_type: "Content-Type" attribute value to set on input blob (ex. "application/json")
        :type content_type: ContentType
        :param encoding: "Content-Encoding" attribute value to set on input blob (ex. "gzip")
        :type encoding: str
        :param average_queue_time: Set average queue time (for internal use)
        :type average_queue_time: float
        :param current_availability: Set current availability (for internal use)
        :type current_availability: str
        """
        if not provider_id and "." in name:
            provider_id = name.split(".")[0]
        self.workspace = workspace
        self.name = name
        self.input_data_format = input_data_format
        self.output_data_format = output_data_format
        self.capability = capability
        self.provider_id = provider_id
        self.content_type = content_type
        self.encoding = encoding
        self._average_queue_time = average_queue_time
        self._current_availability = current_availability

    def __repr__(self):
        return f"<Target name=\"{self.name}\", \
avg. queue time={self._average_queue_time} s, {self._current_availability}>"

    @classmethod
    def from_target_status(
        cls, workspace: "Workspace", status: TargetStatus, **kwargs
    ):
        """Create a Target instance from a given workspace and target status.

        :param workspace: Associated workspace
        :type workspace: Workspace
        :param status: Target status with availability and current queue time
        :type status: TargetStatus
        :return: Target instance
        :rtype: Target
        """
        return cls(
            workspace=workspace,
            name=status.id,
            average_queue_time=status.average_queue_time,
            current_availability=status.current_availability,
            **kwargs
        )

    @classmethod
    def _get_job_class(cls) -> Type[Job]:
        """
        Returns the job class associated to this target.

        The job class used by submit and get_job.  The default is Job.
        """
        return Job
    
    @classmethod
    def _can_send_shots_input_param(cls) -> bool:
        """
        Tells if provider's target class is able to specify shots number for its jobs.
        """
        return cls._SHOTS_PARAM_NAME is not None

    def refresh(self):
        """Update the target availability and queue time"""
        targets = self.workspace._get_target_status(self.name, self.provider_id)
        if len(targets) > 0:
            _, target_status = targets[0]
            self._current_availability = target_status.current_availability
            self._average_queue_time = target_status.average_queue_time
        else:
            raise ValueError(
                f"Cannot refresh the Target status: \
target '{self.name}' of provider '{self.provider_id}' not found."
            )

    @property
    def current_availability(self):
        """
        Current availability.
        """

        return self._current_availability

    @property
    def average_queue_time(self):
        """
        Average queue time.
        """
        
        return self._average_queue_time

    @staticmethod
    def _encode_input_data(data: Any) -> bytes:
        """Encode input data to bytes.
        If the data is already in bytes format, return it.

        :param data: Input data
        :type data: Any
        :return: Encoded input data
        :rtype: bytes
        """
        if isinstance(data, bytes):
            return data
        else:
            stream = io.BytesIO()
            if isinstance(data, dict):
                data = json.dumps(data)
            stream.write(data.encode())
            return stream.getvalue()
        
    def _qir_output_data_format(self) -> str:
        """"Fallback output data format in case of QIR job submission."""
        return "microsoft.quantum-results.v1"

    def submit(
        self,
        input_data: Any,
        name: str = "azure-quantum-job",
        shots: int = None,
        input_params: Union[Dict[str, Any], InputParams, None] = None,
        **kwargs
    ) -> Job:
        """Submit input data and return Job.

        Provide input_data_format, output_data_format and content_type
        keyword arguments to override default values.

        :param input_data: Input data
        :type input_data: Any
        :param name: Job name
        :type name: str
        :param shots: Number of shots, defaults to None
        :type shots: int
        :param input_params: Input parameters
        :type input_params: Dict[str, Any]
        :return: Azure Quantum job
        :rtype: azure.quantum.job.Job
        """

        if isinstance(input_params, InputParams):
            input_params = input_params.as_dict()
        else:
            input_params = input_params or {}
        input_data_format = None
        output_data_format = None
        content_type = None

        # If the input_data is `QirRepresentable`
        # we need to convert it to QIR bitcode and set the necessary parameters for a QIR job.
        if input_data and isinstance(input_data, QirRepresentable):
            input_data_format = kwargs.pop("input_data_format", "qir.v1")
            output_data_format = kwargs.pop("output_data_format", self._qir_output_data_format())
            content_type = kwargs.pop("content_type", "qir.v1")
            # setting UserAgent header to indicate Q# submission
            # TODO: this is a temporary solution. We should be setting the User-Agent header
            # on per-job basis as targets of different types could be submitted using the same Workspace object
            self.workspace.append_user_agent(self._QSHARP_USER_AGENT)

            def _get_entrypoint(input_data):
                # TODO: this method should be part of QirRepresentable protocol
                # and will later move to the QSharpCallable class in the qsharp package
                import re
                method_name = re.search(r"(?:^|\.)([^.]*)$", input_data._name).group(1)
                return f'ENTRYPOINT__{method_name}'

            input_params["entryPoint"] = input_params.get("entryPoint", _get_entrypoint(input_data))
            input_params["arguments"] = input_params.get("arguments", [])
            targetCapability = input_params.get("targetCapability", kwargs.pop("target_capability", self.capability))
            if targetCapability:
                input_params["targetCapability"] = targetCapability
            input_data = input_data._repr_qir_(target=self.name, target_capability=targetCapability)
        else:
            input_data_format = kwargs.pop("input_data_format", self.input_data_format)
            output_data_format = kwargs.pop("output_data_format", self.output_data_format)
            content_type = kwargs.pop("content_type", self.content_type)
            # re-setting UserAgent header to None for passthrough
            self.workspace.append_user_agent(None)
        
        # Set shots number, if possible.
        if self._can_send_shots_input_param():
            input_params_shots = input_params.pop(self.__class__._SHOTS_PARAM_NAME, None)

            # If there is a parameter conflict, choose 'shots'.
            if shots is not None and input_params_shots is not None:
                warnings.warn(
                    f"Parameter 'shots' conflicts with the '{self.__class__._SHOTS_PARAM_NAME}' field of the 'input_params' "
                    "parameter. Please, provide only one option for setting shots. Defaulting to 'shots' parameter."
                )
                final_shots = shots
            
            # The 'shots' parameter has highest priority.
            elif shots is not None:
                final_shots = shots
            # if 'shots' parameter is not specified, try a provider-specific option.
            elif input_params_shots is not None:
                warnings.warn(
                    f"Field '{self.__class__._SHOTS_PARAM_NAME}' from the 'input_params' parameter is subject to change in future versions. "
                    "Please, use 'shots' parameter instead."
                )
                final_shots = input_params_shots
            else:
                final_shots = None
            
            if final_shots is not None:
                input_params[self.__class__._SHOTS_PARAM_NAME] = final_shots

        encoding = kwargs.pop("encoding", self.encoding)
        blob = self._encode_input_data(data=input_data)
        job_cls = type(self)._get_job_class()
        return job_cls.from_input_data(
            workspace=self.workspace,
            name=name,
            target=self.name,
            input_data=blob,
            content_type=content_type,
            encoding=encoding,
            provider_id=self.provider_id,
            input_data_format=input_data_format,
            output_data_format=output_data_format,
            input_params=input_params,
            session_id=self.get_latest_session_id(),
            **kwargs
        )

    def make_params(self):
        """
        Returns an input parameter object for convenient creation of input
        parameters.
        """
        return InputParams()

    def estimate_cost(
        self,
        input_data: Any,
        input_params: Union[Dict[str, Any], None] = None
    ):
        """
        Estimate the cost for a given circuit.
        """
        return NotImplementedError("Price estimation is not implemented yet for this target.")

    def _get_azure_workspace(self) -> "Workspace":
        return self.workspace

    def _get_azure_target_id(self) -> str:
        return self.name

    def _get_azure_provider_id(self) -> str:
        return self.provider_id


def _determine_shots_or_deprecated_num_shots(
    shots: int = None,
    num_shots: int = None,
) -> int:
    """
    This helper function checks if the deprecated 'num_shots' parameter is specified.
    In earlier versions it was possible to pass this parameter to specify shots number for a job,
    but now we only check for it for compatibility reasons.  
    """
    final_shots = None
    if shots is not None and num_shots is not None:
        warnings.warn(
            "Both 'shots' and 'num_shots' parameters were specified. Defaulting to 'shots' parameter. "
            "Please, use 'shots' since 'num_shots' will be deprecated.",
            category=DeprecationWarning,
        )
        final_shots = shots
        
    elif shots is not None:
        final_shots = shots
    elif num_shots is not None:
        warnings.warn(
            "The 'num_shots' parameter will be deprecated. Please, use 'shots' parameter instead.",
            category=DeprecationWarning,
        )
        final_shots = num_shots

    return final_shots


##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

import logging
import uuid
import io
import gzip
import json
import threading
import sys

from typing import List, Union, Dict, Optional
from azure.quantum import Workspace
from azure.quantum.optimization import Term, Problem, ProblemType
from azure.quantum.storage import (
    StreamedBlob,
    ContainerClient,
    BlobClient,
    download_blob,
)
from queue import Queue, Empty

logger = logging.getLogger(__name__)

__all__ = ["StreamingProblem"]


class StreamingProblem(object):
    """Problem to be streamed to the service.

    Streaming problems are uploaded on the fly as terms are added,
    meaning that the whole problem representation is not kept in memory. This
    is very useful when constructing large  problems.

    :param workspace: Workspace to upload problem to
    :type workspace: Workspace
    :param name: Problem name
    :type name: str
    :param terms: Problem terms, depending on solver. Defaults to None
    :type terms: Optional[List[TermBase]], optional
    :param init_config: Optional configuration details,
     depending on solver. Defaults to None
    :type init_config: Optional[Dict[str,int]], optional
    :param problem_type: Problem type (ProblemType.pubo or
     ProblemType.ising), defaults to ProblemType.ising
    :type problem_type: ProblemType, optional
    """

    def __init__(
        self,
        workspace: Workspace,
        name: str = "Optimization Problem",
        terms: Optional[List[Term]] = None,
        init_config: Optional[Dict[str, int]] = None,
        problem_type: ProblemType = ProblemType.ising,
        metadata: Dict[str, str] = {},
        **kw,
    ):
        super(StreamingProblem, self).__init__(**kw)
        self.name = name
        self._id = str(uuid.uuid1())
        self.workspace = workspace
        self.problem_type = problem_type
        self.init_config = init_config
        self.terms_queue = Queue()
        self.uploaded_uri = None
        self.upload_to_url = None
        self.uploader = None
        self.__n_couplers = 0
        self.stats = {
            "type": problem_type.name,
            "max_coupling": 0,
            "avg_coupling": 0,
            "min_coupling": sys.maxsize,
            "num_terms": 0,
        }
        self.upload_size_threshold = 10e6
        self.upload_terms_threshold = 1000
        self.metadata = metadata
        if terms is not None and len(terms) > 0:
            self.add_terms(terms.copy())

    @property
    def id(self):
        return self._id

    def add_term(self, c: Union[int, float], indices: List[int]):
        """Adds a single monomial term to the `Problem`
        representation and queues it to be uploaded

        :param c: The cost or weight of this term
        :type c: int, float
        :param indices: The variable indices that are in this term
        :type indices: List[int]
        """
        self.add_terms([Term(indices=indices, c=c)])

    def _get_upload_coords(self):
        blob_name = self.id
        if self.upload_to_url:
            blob_client = BlobClient.from_blob_url(self.upload_to_url)
            container_client = ContainerClient.from_container_url(
                self.workspace._get_linked_storage_sas_uri(
                    blob_client.container_name
                )
            )
            blob_name = blob_client.blob_name
        elif not self.workspace.storage:
            # No storage account is passed, use the linked one
            container_uri = self.workspace._get_linked_storage_sas_uri(self.id)
            container_client = ContainerClient.from_container_url(
                container_uri
            )
        else:
            # Use the specified storage account
            container_client = ContainerClient.from_connection_string(
                self.workspace.storage, self.id
            )

        return {"blob_name": blob_name, "container_client": container_client}

    def add_terms(
        self,
        terms: List[Term]
    ):
        """Adds a list of terms to the `Problem`
         representation and queues them to be uploaded. Special terms are not supported.

        :param terms: The list of terms to add to the problem
        """
        if self.uploaded_uri is not None:
            raise Exception("Cannot add terms after problem has been uploaded")

        if terms is not None:
            if self.uploader is None:
                upload_coords = self._get_upload_coords()
                self.uploader = JsonStreamingProblemUploader(
                    problem=self,
                    container=upload_coords["container_client"],
                    name=upload_coords["blob_name"],
                    upload_size_threshold=self.upload_size_threshold,
                    upload_term_threshold=self.upload_terms_threshold,
                )
                self.uploader.start()
            elif self.uploader.is_done():
                raise Exception(
                    "Cannot add terms after problem has been uploaded"
                )

            max_coupling = -sys.float_info.max
            min_coupling = sys.float_info.max
            for term in terms:
                if isinstance(term, Term):
                    n = len(term.ids)
                    max_coupling = max(max_coupling, n)
                    min_coupling = min(min_coupling, n)
                    self.__n_couplers += n
                    self.stats["num_terms"] += 1
                else:
                    raise Exception(
                        "Unsupported statistics in streamingproblem for TermBase subclass {}.".format(type(term))
                    )
                        
            self.stats["avg_coupling"] = (
                self.__n_couplers / self.stats["num_terms"]
            )
            if self.stats["max_coupling"] < max_coupling:
                self.stats["max_coupling"] = max_coupling
            if self.stats["min_coupling"] > min_coupling:
                self.stats["min_coupling"] = min_coupling
            self.terms_queue.put(terms)

    def download(self):
        """Downloads the uploaded problem as an instance of `Problem`"""
        if not self.uploaded_uri:
            raise Exception(
                "StreamingProblem may not be downloaded before it is uploaded"
            )

        coords = self._get_upload_coords()
        blob = coords["container_client"].get_blob_client(coords["blob_name"])
        contents = download_blob(blob.url)
        return Problem.deserialize(contents, self.name)

    def upload(
        self,
        workspace,
        container_name: str = "optimization-problems",
        blob_name: str = None,
    ):
        """Uploads an optimization problem instance
           to the cloud storage linked with the Workspace.

        :param workspace: interaction terms of the problem.
        :return: uri of the uploaded problem
        """
        if not self.uploaded_uri:
            self.uploader.blob_properties = {
                k: str(v) for k, v in {**self.stats, **self.metadata}.items()
            }
            self.terms_queue.put(None)
            blob = self.uploader.join()
            self.uploaded_uri = blob.getUri(not not self.workspace.storage)
            self.uploader = None
            self.terms_queue = None

        return self.uploaded_uri


class JsonStreamingProblemUploader:
    """Helper class for uploading json problem files in chunks.

    :param problem: Back-ref to the problem being uploaded
    :param container: Reference to the container
     client in which to store the problem
    :param name: Name of the problem (added to blob metadata)
    :param upload_size_threshold: Chunking threshold (in bytes).
     Once the internal buffer reaches this size, the chunk will be uploaded.
    :param upload_term_threshold: Chunking threshold (in terms).
     Once this many terms are ready to be uploaded, the chunk will be uploaded.
    :param blob_properties: Properties to set on the blob.
    """

    def __init__(
        self,
        problem: StreamingProblem,
        container: ContainerClient,
        name: str,
        upload_size_threshold: int,
        upload_term_threshold: int,
        blob_properties: Dict[str, str] = None,
    ):
        self.problem = problem
        self.started_upload = False
        self.blob = StreamedBlob(
            container,
            name,
            "application/json",
            self._get_content_type(),
        )
        self.compressedStream = io.BytesIO()
        self.compressor = (
            gzip.GzipFile(mode="wb", fileobj=self.compressedStream)
        )
        self.uploaded_terms = 0
        self.blob_properties = blob_properties
        self.__thread = None
        self.__queue_wait_timeout = 1
        self.__upload_terms_threshold = upload_term_threshold
        self.__upload_size_threshold = upload_size_threshold
        self.__read_pos = 0

    def _get_content_type(self):
        return "gzip"

    def start(self):
        """Starts the problem uploader in another thread"""
        if self.__thread is not None:
            raise Exception(
                "JsonStreamingProblemUploader thread already started"
            )

        self.__thread = threading.Thread(target=self._run_queue)
        self.__thread.start()

    def join(self, timeout: float = None) -> StreamedBlob:
        """Joins the problem uploader thread -
        returning when it completes or when `timeout` is hit

        :param timeout: The the time to wait for the thread to complete.
        If omitted, the method will wait until the thread completes
        """
        if self.__thread is None:
            raise Exception("JsonStreamingProblemUploader has not started")

        self.__thread.join(timeout=timeout)
        return self.blob

    def is_done(self):
        """True if the thread uploader has completed"""
        return not self.__thread.is_alive()

    def _run_queue(self):
        continue_processing = True
        terms = []
        while continue_processing:
            try:
                new_terms = self.problem.terms_queue.get(
                    block=True, timeout=self.__queue_wait_timeout
                )
                if new_terms is None:
                    continue_processing = False
                else:
                    terms = terms + new_terms
                    if len(terms) < self.__upload_terms_threshold:
                        continue
            except Empty:
                pass
            except Exception as e:
                raise e

            if len(terms) > 0:
                self._upload_next(terms)
                terms = []

        self._finish_upload()

    def _upload_start(self, terms):
        self.started_upload = True
        self._upload_chunk(
            f'{{"cost_function":{{"version":"{self._get_version()}",'
            + f'"type":"{self._scrub(self.problem.problem_type.name)}",'
            + self._get_initial_config_string()
            + '"terms":['
            + self._get_terms_string(terms)
        )

    def _get_initial_config_string(self):
        if self.problem.init_config:
            return (
                f'{"initial_configuration":}'
                + json.dumps(self.problem.init_config)
                + ","
            )
        return ""

    def _get_version(self):
        return "1.1" if self.problem.init_config else "1.0"

    def _get_terms_string(self, terms):
        result = ("," if self.uploaded_terms > 0 else "") + ",".join(
            [json.dumps(term.to_dict()) for term in terms]
        )
        self.uploaded_terms += len(terms)
        return result

    def _scrub(self, s):
        if '"' in s:
            raise "string should not contain a literal double quote '\"'"

        return s

    def _upload_next(self, terms):
        if not self.started_upload:
            self._upload_start(terms)
        else:
            self._upload_chunk(self._get_terms_string(terms))

    def _maybe_compress_bits(self, chunk: bytes, is_final: bool):
        if self.compressor is None:
            return chunk

        if self.__read_pos > 0:
            self.compressedStream.truncate(0)
            self.compressedStream.seek(0)

        self.compressor.write(chunk)
        if is_final:
            self.compressor.flush()
            self.compressor.close()
        elif (
            self.compressedStream.getbuffer().nbytes
            < self.__upload_size_threshold
        ):
            self.__read_pos = 0
            return

        self.compressedStream.seek(0)
        compressed = self.compressedStream.read(-1)
        self.__read_pos = 0 if compressed is None else len(compressed)
        return compressed

    def _upload_chunk(self, chunk: str, is_final: bool = False):
        compressed = self._maybe_compress_bits(chunk.encode(), is_final)
        if compressed is None:
            return
        if len(compressed) > 0:
            self.blob.upload_data(compressed)

    def _finish_upload(self):
        if not self.started_upload:
            self._upload_start([])

        self._upload_chunk(f'{"]}}"}', True)
        self.blob.commit(metadata=self.blob_properties)



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
import logging
import os
import requests
import re

from azure.identity._constants import EnvironmentVariables
from azure.identity._internal import get_default_authority, normalize_authority
from azure.identity import (
    AzurePowerShellCredential,
    EnvironmentCredential,
    ManagedIdentityCredential,
    SharedTokenCacheCredential,
    AzureCliCredential,
    VisualStudioCodeCredential,
    InteractiveBrowserCredential,
    DeviceCodeCredential,
    _credentials
)
from ._chained import _ChainedTokenCredential
from ._token import _TokenFileCredential

try:
    from typing import TYPE_CHECKING
except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
    from typing import Any, List
    from azure.core.credentials import AccessToken, TokenCredential

_LOGGER = logging.getLogger(__name__)


class _DefaultAzureCredential(_ChainedTokenCredential):
    """
    Based on Azure.Identity.DefaultAzureCredential from:
    https://github.com/Azure/azure-sdk-for-python/blob/master/sdk/identity/azure-identity/azure/identity/_credentials/default.py

    The three key differences are:
    1) Inherit from _ChainedTokenCredential, which has 
       more aggressive error handling than ChainedTokenCredential
    2) Instantiate the internal credentials the first time the get_token gets called
       such that we can get the tenant_id if it was not passed by the user (but we don't
       want to do that in the constructor).
       We automatically identify the user's tenant_id for a given subscription 
       so that users with MSA accounts don't need to pass it.
       This is a mitigation for bug https://github.com/Azure/azure-sdk-for-python/issues/18975
       We need the following parameters to enable auto-detection of tenant_id
       - subscription_id
       - arm_base_url (defaults to the production url "https://management.azure.com/")
    3) Add custom TokenFileCredential as first method to attempt, which will look for a local access token.
    """
    def __init__(self, **kwargs):
        # type: (**Any) -> None
        self._successful_tenant_id = None

        self.authority = kwargs.pop("authority", None)
        self.authority = normalize_authority(self.authority) if self.authority else get_default_authority()

        self.interactive_browser_tenant_id = kwargs.pop(
            "interactive_browser_tenant_id", os.environ.get(EnvironmentVariables.AZURE_TENANT_ID)
        )

        self.subscription_id = kwargs.pop(
            "subscription_id", os.environ.get("SUBSCRIPTION_ID")
        )
        self.arm_base_url = kwargs.pop(
            "arm_base_url", "https://management.azure.com/"
        )

        self.managed_identity_client_id = kwargs.pop(
            "managed_identity_client_id", os.environ.get(EnvironmentVariables.AZURE_CLIENT_ID)
        )

        self.shared_cache_username = kwargs.pop("shared_cache_username", os.environ.get(EnvironmentVariables.AZURE_USERNAME))
        self.shared_cache_tenant_id = kwargs.pop(
            "shared_cache_tenant_id", os.environ.get(EnvironmentVariables.AZURE_TENANT_ID)
        )

        self.vscode_tenant_id = kwargs.pop(
            "visual_studio_code_tenant_id", os.environ.get(EnvironmentVariables.AZURE_TENANT_ID)
        )

        self.exclude_token_file_credential = kwargs.pop("exclude_token_file_credential", False)
        self.exclude_environment_credential = kwargs.pop("exclude_environment_credential", False)
        self.exclude_managed_identity_credential = kwargs.pop("exclude_managed_identity_credential", False)
        self.exclude_shared_token_cache_credential = kwargs.pop("exclude_shared_token_cache_credential", True)
        self.exclude_visual_studio_code_credential = kwargs.pop("exclude_visual_studio_code_credential", False)
        self.exclude_cli_credential = kwargs.pop("exclude_cli_credential", False)
        self.exclude_interactive_browser_credential = kwargs.pop("exclude_interactive_browser_credential", True)
        self.exclude_device_code_credential = kwargs.pop("exclude_device_code_credential", False)
        self.exclude_powershell_credential = kwargs.pop("exclude_powershell_credential", False)

        # credentials will be created lazy on the first call to get_token
        super(_DefaultAzureCredential, self).__init__()

    def _initialize_credentials(self):
        if self.subscription_id is not None \
           and self.arm_base_url is not None:
            if self.vscode_tenant_id is None:
                self.vscode_tenant_id = self._get_tenant_id(arm_base_url=self.arm_base_url, subscription_id=self.subscription_id)
            if self.shared_cache_tenant_id is None:
                self.shared_cache_tenant_id = self._get_tenant_id(arm_base_url=self.arm_base_url, subscription_id=self.subscription_id)
            if self.interactive_browser_tenant_id is None:
                self.interactive_browser_tenant_id = self._get_tenant_id(arm_base_url=self.arm_base_url, subscription_id=self.subscription_id)

        credentials = []  # type: List[TokenCredential]
        if not self.exclude_token_file_credential:
            credentials.append(_TokenFileCredential())
        if not self.exclude_environment_credential:
            credentials.append(EnvironmentCredential(authority=self.authority))
        if not self.exclude_managed_identity_credential:
            credentials.append(ManagedIdentityCredential(client_id=self.managed_identity_client_id))
        if not self.exclude_shared_token_cache_credential and SharedTokenCacheCredential.supported():
            try:
                # username and/or tenant_id are only required when the cache contains tokens for multiple identities
                shared_cache = SharedTokenCacheCredential(
                    username=self.shared_cache_username, tenant_id=self.shared_cache_tenant_id, authority=self.authority
                )
                credentials.append(shared_cache)
            except Exception as ex:  # pylint:disable=broad-except
                _LOGGER.info("Shared token cache is unavailable: '%s'", ex)
        if not self.exclude_visual_studio_code_credential:
            credentials.append(VisualStudioCodeCredential(tenant_id=self.vscode_tenant_id))
        if not self.exclude_cli_credential:
            credentials.append(AzureCliCredential())
        if not self.exclude_powershell_credential:
            credentials.append(AzurePowerShellCredential())
        if not self.exclude_interactive_browser_credential:
            credentials.append(InteractiveBrowserCredential(tenant_id=self.interactive_browser_tenant_id))
        if not self.exclude_device_code_credential:
            credentials.append(DeviceCodeCredential(tenant_id=self.interactive_browser_tenant_id))

        self.credentials = credentials

    def get_token(self, *scopes, **kwargs):
        # type: (*str, **Any) -> AccessToken
        """Request an access token for `scopes`.
        This method is called automatically by Azure SDK clients.
        :param str scopes: desired scopes for the access token. This method requires at least one scope.
        :raises ~azure.core.exceptions.ClientAuthenticationError: authentication failed. The exception has a
          `message` attribute listing each authentication attempt and its error message.
        """
        # add credentials the first time it runs the get_token
        # such that the _get_tenant_id can be called only when needed
        if self.credentials is None \
           or len(self.credentials) == 0:
            self._initialize_credentials()

        return super(_DefaultAzureCredential, self).get_token(*scopes, **kwargs)

    def _get_tenant_id(self, arm_base_url:str, subscription_id:str):
        if arm_base_url is None:
            raise ValueError("arm_base_url is mandatory parameter")
        if subscription_id is None:
            raise ValueError("subscription_id is mandatory parameter")

        # returns the cached tenant_id if available
        if self._successful_tenant_id is not None:
            return self._successful_tenant_id

        try:
            uri = (
                f"{arm_base_url}/subscriptions/"
                + f"{subscription_id}?api-version=2018-01-01"
            )
            response = requests.get(uri)

            # This gnarly piece of code is how we get the guest tenant
            # authority associated with the subscription.
            # We make a unauthenticated request to ARM and extract the tenant
            # authority from the WWW-Authenticate header in the response.
            # The header is of the form:
            # Bearer authorization_uri=
            # https://login.microsoftonline.com/tenantId, key1=value1s
            auth_header = response.headers["WWW-Authenticate"]
            _LOGGER.debug(
                f'{"got the following auth header from"}'
                f"the management endpoint: {auth_header}"
            )

            trimmed_auth_header = auth_header[
                len("Bearer "):
            ]  # trim the leading 'Bearer '
            trimmed_auth_header_parts = trimmed_auth_header.split(
                ","
            )  # get the various k=v parts
            key_value_pairs = dict(
                map(lambda s: tuple(s.split("=")), trimmed_auth_header_parts)
            )  # make the parts into a dictionary
            quoted_tenant_uri = key_value_pairs[
                "authorization_uri"
            ]  # get the value of the 'authorization_uri' key
            tenant_uri = quoted_tenant_uri[
                1:-1
            ]  # strip it of surrounding quotes

            _LOGGER.debug(
                f'{"got the following tenant uri from"}'
                + f"the authentication header: {tenant_uri}"
            )

            regex = re.compile(pattern=r"([a-f0-9]+[-]){4}[a-f0-9]+", flags=re.IGNORECASE)
            self._successful_tenant_id = regex.search(tenant_uri).group()
            return self._successful_tenant_id

        except Exception as e:
            _LOGGER.debug(
                f"Failed to get tenant authority for subscription: {e}"
            )
            return None


##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
__all__ = ['MicrosoftEstimatorResult']

from typing import Any, Dict, List, Optional, Union

import json
import markdown


class HTMLWrapper:
    """
    Simple HTML wrapper to expose _repr_html_ for Jupyter clients.
    """
    def __init__(self, content: str):
        self.content = content

    def _repr_html_(self):
        return self.content

class MicrosoftEstimatorResult(dict):
    """
    Microsoft Resource Estimator result.

    Job results from the `microsoft.estimator` target are represented by
    instances of this class.  The class represents simple resource estimation
    results as well as batching resource estimation results.  The latter can
    be indexed by an integer index to access an individual result from the
    batching result.
    """
    MAX_DEFAULT_ITEMS_IN_TABLE = 5

    def __init__(self, data: Union[Dict, List]):
        self._data = data

        if isinstance(data, dict):
            super().__init__(data)

            self._is_simple = True
            if MicrosoftEstimatorResult._is_succeeded(self):
                self._repr = self._item_result_table()
                self.summary = HTMLWrapper(self._item_result_summary_table())
                self.diagram = EstimatorResultDiagram(self.data().copy())

        elif isinstance(data, list):
            super().__init__({idx: MicrosoftEstimatorResult(item_data)
                              for idx, item_data in enumerate(data)})

            self._data = data
            self._is_simple = False
            num_items = len(data)
            self._repr = ""
            if num_items > self.MAX_DEFAULT_ITEMS_IN_TABLE:
                self._repr += "<p><b>Info:</b> <i>The overview table is " \
                              "cut off after " \
                              f"{self.MAX_DEFAULT_ITEMS_IN_TABLE} items. If " \
                              "you want to see all items, suffix the result " \
                              "variable with <code>[:]</code></i></p>"
                num_items = self.MAX_DEFAULT_ITEMS_IN_TABLE
            self._repr += self._batch_result_table(range(num_items))

            # Add plot function for batching jobs
            self.plot = self._plot
            self.summary_data_frame = self._summary_data_frame
    
    def _is_succeeded(self):
        return 'status' in self and self['status'] == "success"

    def data(self, idx: Optional[int] = None) -> Any:
        """
        Returns raw data of the result object.

        In case of a batching job, you can pass an index to access a specific
        item.
        """
        if idx is None:
            return self._data
        elif not self._is_simple:
            return self._data[idx]
        else:
            msg = "Cannot pass parameter 'idx' to 'data' for non-batching job"
            raise ValueError(msg)

    def _repr_html_(self):
        """
        HTML table representation of the result.
        """
        return self._repr

    def __getitem__(self, key):
        """
        If the result represents a batching job and key is a slice, a
        side-by-side table comparison is shown for the indexes represented by
        the slice.

        Otherwise, the key is used to access the raw data directly.
        """
        if isinstance(key, slice):
            if self._is_simple:
                msg = "Cannot pass slice to '__getitem__' for non-batching job"
                raise ValueError(msg)
            return HTMLWrapper(self._batch_result_table(range(len(self))[key]))
        else:
            return super().__getitem__(key)

    def _plot(self, **kwargs):
        """
        Plots all result items in a space time plot, where the x-axis shows
        total runtime, and the y-axis shows total number of physical qubits.
        Both axes are in log-scale.
        Attributes:
            labels (list): List of labels for the legend.
        """
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise ImportError(
                "Missing optional 'matplotlib' dependency. To install run: "
                "pip install matplotlib"
            )

        labels = kwargs.pop("labels", [])

        [xs, ys] = zip(*[
            (self.data(i)['physicalCounts']['runtime'],
             self.data(i)['physicalCounts']['physicalQubits'])
            for i in range(len(self))])

        _ = plt.figure(figsize=(15, 8))

        plt.ylabel('Physical qubits')
        plt.xlabel('Runtime')
        plt.loglog()
        for i, (x, y) in enumerate(zip(xs, ys)):
            if isinstance(labels, list) and i < len(labels):
                label = labels[i]
            else:
                label = str(i)
            plt.scatter(x=[x], y=[y], label=label, marker="os+x"[i % 4])

        nsec = 1
        usec = 1e3 * nsec
        msec = 1e3 * usec
        sec = 1e3 * msec
        min = 60 * sec
        hour = 60 * min
        day = 24 * hour
        week = 7 * day
        month = 31 * day
        year = 365 * month
        decade = 10 * year
        century = 10 * decade

        time_units = [
            nsec, usec, msec, sec, min, hour, day, week,
            month, year, decade, century]
        time_labels = [
            "1 ns", "1 µs", "1 ms", "1 s", "1 min", "1 hour", "1 day",
            "1 week", "1 month", "1 year", "1 decade", "1 century"]

        cutoff = next(
            (i for i, x in enumerate(time_units) if x > max(xs)),
            len(time_units) - 1) + 1

        plt.xticks(time_units[0:cutoff], time_labels[0:cutoff], rotation=90)
        plt.legend(loc="upper left")
        plt.show()

    @property
    def call_graph(self):
        """
        Shows the call graph of a simple resource estimation result with
        profiling information.
        """
        try:
            import graphviz
        except ImportError:
            raise ImportError(
                "Missing optional 'graphviz' dependency. To install run: "
                "pip install graphviz"
            )

        if not self._is_simple:
            raise ValueError("The `call_graph` method cannot be called on a "
                             "batching result, try indexing into the result "
                             "first")

        if not hasattr(self, "_call_graph"):
            from itertools import groupby

            data = self.data().get("callGraph", None)

            if data is None:
                raise ValueError("The result does not contain any profiling "
                                 "information. Set "
                                 "`profiling.call_stack_depth` to some value")

            g = graphviz.Digraph()
            g.attr('node', shape='box', style='rounded, filled',
                   fontname='Arial', fontsize='10', margin='0.05,0.05',
                   height='0', width='0', fillcolor='#f6f6f6', color='#e3e3e3')
            g.attr('edge', color='#d0d0d0')

            nodes_indexed = [{**node, 'index': index}
                             for index, node in enumerate(data['nodes'])]

            def sorter(node): return node['depth']
            nodes_indexed.sort(key=sorter)
            for _, nodes in groupby(nodes_indexed, sorter):
                with g.subgraph() as s:
                    s.attr(rank='same')
                    for node in nodes:
                        s.node(str(node['index']), node['name'])

            for edge in data['edges']:
                g.edge(str(edge[0]), str(edge[1]))

            self._call_graph = g

        return self._call_graph

    @property
    def profile(self):
        """
        """
        if not self._is_simple:
            raise ValueError("The `call_graph` method cannot be called on a "
                             "batching result, try indexing into the result "
                             "first")

        if not hasattr(self, "_profile"):
            import base64
            import json

            profile = self.data().get("profile", None)

            if profile is None:
                raise ValueError("The result does not contain any profiling "
                                 "information. Set "
                                 "`profiling.call_stack_depth` to some value")

            profile_encoded = json.dumps(profile).encode('utf-8')
            data64 = base64.b64encode(profile_encoded).decode('utf-8')

            self._profile = f"""
            <a href="data:text/json;base64,{data64}" download="profile.json">
            Download the profile</a> to your computer.  Then open the profile
            by dragging it into <a href="https://speedscope.app"
            target="_blank">speedscope</a>.
            """

        return HTMLWrapper(self._profile)

    @property
    def json(self):
        """
        Returns a JSON representation of the resource estimation result data.
        """
        if not hasattr(self, "_json"):
            import json
            self._json = json.dumps(self._data)

        return self._json

    def _summary_data_frame(self, **kwargs):
        try:
            import pandas as pd
        except ImportError:
            raise ImportError(
                "Missing optional 'pandas' dependency. To install run: "
                "pip install pandas"
            )

        # get labels or use default value, then extend with missing elements,
        # and truncate extra elements
        labels = kwargs.pop("labels", [])
        labels.extend(range(len(labels), len(self)))
        labels = labels[:len(self)]

        def get_row(result):
            if MicrosoftEstimatorResult._is_succeeded(result):
                formatted = result["physicalCountsFormatted"]

                return (
                    formatted["algorithmicLogicalQubits"],
                    formatted["logicalDepth"],
                    formatted["numTstates"],
                    result["logicalQubit"]["codeDistance"],
                    formatted["numTfactories"],
                    formatted["physicalQubitsForTfactoriesPercentage"],
                    formatted["physicalQubits"],
                    formatted["rqops"],
                    formatted["runtime"]
                )
            else:
                return ['No solution found'] * 9

        data = [get_row(self.data(index)) for index in range(len(self))]
        columns = ["Logical qubits", "Logical depth", "T states",
                   "Code distance", "T factories", "T factory fraction",
                   "Physical qubits", "rQOPS", "Physical runtime"]
        return pd.DataFrame(data, columns=columns, index=labels)

    def _item_result_table(self):
        html = ""

        md = markdown.Markdown(extensions=['mdx_math'])
        for group in self['reportData']['groups']:
            html += f"""
                <details {"open" if group['alwaysVisible'] else ""}>
                    <summary style="display:list-item">
                        <strong>{group['title']}</strong>
                    </summary>
                    <table>"""
            for entry in group['entries']:
                val = self
                for key in entry['path'].split("/"):
                    val = val[key]
                explanation = md.convert(entry["explanation"])
                html += f"""
                    <tr>
                        <td style="font-weight: bold; vertical-align: top; white-space: nowrap">{entry['label']}</td>
                        <td style="vertical-align: top; white-space: nowrap">{val}</td>
                        <td style="text-align: left">
                            <strong>{entry["description"]}</strong>
                            <hr style="margin-top: 2px; margin-bottom: 0px; border-top: solid 1px black" />
                            {explanation}
                        </td>
                    </tr>
                """
            html += "</table></details>"

        html += f"<details><summary style=\"display:list-item\"><strong>Assumptions</strong></summary><ul>"
        for assumption in self['reportData']['assumptions']:
            html += f"<li>{md.convert(assumption)}</li>"
        html += "</ul></details>"

        return html

    def _item_result_summary_table(self):
        html = """
            <style>
                .aqre-tooltip {
                    position: relative;
                    border-bottom: 1px dotted black;
                }

                .aqre-tooltip .aqre-tooltiptext {
                    font-weight: normal;
                    visibility: hidden;
                    width: 600px;
                    background-color: #e0e0e0;
                    color: black;
                    text-align: center;
                    border-radius: 6px;
                    padding: 5px 5px;
                    position: absolute;
                    z-index: 1;
                    top: 150%;
                    left: 50%;
                    margin-left: -200px;
                    border: solid 1px black;
                }

                .aqre-tooltip .aqre-tooltiptext::after {
                    content: "";
                    position: absolute;
                    bottom: 100%;
                    left: 50%;
                    margin-left: -5px;
                    border-width: 5px;
                    border-style: solid;
                    border-color: transparent transparent black transparent;
                }

                .aqre-tooltip:hover .aqre-tooltiptext {
                    visibility: visible;
                }
            </style>"""

        md = markdown.Markdown(extensions=['mdx_math'])
        for group in self['reportData']['groups']:
            html += f"""
                <details {"open" if group['alwaysVisible'] else ""}>
                    <summary style="display:list-item">
                        <strong>{group['title']}</strong>
                    </summary>
                    <table>"""
            for entry in group['entries']:
                val = self
                for key in entry['path'].split("/"):
                    val = val[key]
                explanation = md.convert(entry["explanation"])
                html += f"""
                    <tr class="aqre-tooltip">
                        <td style="font-weight: bold"><span class="aqre-tooltiptext">{explanation}</span>{entry['label']}</td>
                        <td>{val}</td>
                        <td style="text-align: left">{entry["description"]}</td>
                    </tr>
                """
            html += "</table></details>"

        html += f"<details><summary style=\'display:list-item\'><strong>Assumptions</strong></summary><ul>"
        for assumption in self['reportData']['assumptions']:
            html += f"<li>{md.convert(assumption)}</li>"
        html += "</ul></details>"

        return html

    def _batch_result_table(self, indices):
        succeeded_item_indices = [i for i in indices if MicrosoftEstimatorResult._is_succeeded(self[i])]
        if len(succeeded_item_indices) == 0:
            print("None of the jobs succeeded")
            return ""
        
        first_succeeded_item_index = succeeded_item_indices[0]

        html = ""

        md = markdown.Markdown(extensions=['mdx_math'])

        item_headers = "".join(f"<th>{i}</th>" for i in indices)

        for group_index, group in enumerate(self[first_succeeded_item_index]['reportData']['groups']):
            html += f"""
                <details {"open" if group['alwaysVisible'] else ""}>
                    <summary style="display:list-item">
                        <strong>{group['title']}</strong>
                    </summary>
                    <table>
                        <thead><tr><th>Item</th>{item_headers}</tr></thead>"""

            visited_entries = set()

            for entry in [entry for index in succeeded_item_indices for entry in self[index]['reportData']['groups'][group_index]['entries']]:
                label = entry['label']
                if label in visited_entries:
                    continue
                visited_entries.add(label)

                html += f"""
                    <tr>
                        <td style="font-weight: bold; vertical-align: top; white-space: nowrap">{label}</td>
                """

                for index in indices:
                    val = self[index]
                    if index in succeeded_item_indices:
                        for key in entry['path'].split("/"):
                            if key in val:
                                val = val[key]
                            else:
                                val = "N/A"
                                break
                    else:
                        val = "N/A"
                    html += f"""
                            <td style="vertical-align: top; white-space: nowrap">{val}</td>
                    """

                html += """
                    </tr>
                """
            html += "</table></details>"

        html += f"<details><summary style=\"display:list-item\"><strong>Assumptions</strong></summary><ul>"
        for assumption in self[first_succeeded_item_index]['reportData']['assumptions']:
            html += f"<li>{md.convert(assumption)}</li>"
        html += "</ul></details>"

        return html

    @staticmethod
    def _is_succeeded(obj):
        return 'status' in obj and obj['status'] == "success"
    
class EstimatorResultDiagram:
    def __init__(self, data):
        data.pop("reportData")
        self.data_json = json.dumps(data).replace(" ", "")
        self.vis_lib = "https://cdn-aquavisualization-prod.azureedge.net/resource-estimation/index.js"
        self.space = HTMLWrapper(self._space_diagram())
        self.time = HTMLWrapper(self._time_diagram())

    def _space_diagram(self):
        html = f"""
            <script src={self.vis_lib}></script>
            <re-space-diagram data={self.data_json}></re-space-diagram>"""
        return html

    def _time_diagram(self):
        html = f"""
            <script src={self.vis_lib}></script>
            <re-time-diagram data={self.data_json}></re-time-diagram>"""
        return html


"""Defines targets and helper functions for the Pasqal provider"""

##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

__all__ = [
    "Result",
]

import json
from typing import Union, Dict, List, TypeVar, cast

from ...job import Job


class Result:
    """Downloads the data of a completed Job and provides a dictionary of registers.

    .. highlight:: python
    .. code-block::

       from azure.quantum.job import Job
       from azure.quantum.target.pasqal import Result
       job = Job(...)  # This job should come from a Pasqal target
       job.wait_until_completed()
       result = Result(job)

    """

    def __init__(self, job: Job) -> None:
        """
        Decode the results of a Job with output type of "pasqal.pulser-results.v1"

        :param job: Azure Quantum job
        :type job: Job
        :raises: RuntimeError if the job has not completed successfully
        """

        if job.details.status != "Succeeded":
            raise RuntimeError(
                "Cannot retrieve results as job execution failed "
                f"(status: {job.details.status}."
                f"error: {job.details.error_data})"
            )
        self.data = cast(Dict[str, int], json.loads(job.download_data(job.details.output_data_uri)))

    def __getitem__(self, register_name: str) -> int:
        return self.data[register_name]


##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

"""Defines argument types for Microsoft Estimator"""

from .types import EmptyArray, Pauli, Range, Result

__all__ = ['EmptyArray', 'Pauli', 'Range', 'Result']



"""Defines targets and helper functions for the Rigetti provider"""

##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

__all__ = [
    "Readout",
    "Result",
]

import json
from typing import Union, Dict, List, TypeVar, cast

from ...job import Job

RawData = Union[int, float, List[float]]


class Result:
    """Downloads the data of a completed Job and extracts the ``Readout`` for each register.

    >>> from azure.quantum.job import Job
    >>> from azure.quantum.target.rigetti import Result
    >>> job = Job(...)  # This job should come from a Rigetti target
    >>> job.wait_until_completed()
    >>> result = Result(job)
    >>> ro_data = result["ro"]
    >>> first_shot_data = ro_data[0]
    """

    def __init__(self, job: Job) -> None:
        """
        Decode the results of a Job with output type of "rigetti.quil-results.v1"

        :param job: Azure Quantum Job
        :type job: Job
        :raises: RuntimeError if the job has not completed successfully
        """

        if job.details.status != "Succeeded":
            raise RuntimeError(
                "Cannot retrieve results as job execution failed "
                f"(status: {job.details.status}."
                f"error: {job.details.error_data})"
            )
        data = cast(Dict[str, List[List[RawData]]], json.loads(job.download_data(job.details.output_data_uri)))
        self.data_per_register: Dict[str, Readout] = {k: create_readout(v) for k, v in data.items()}

    def __getitem__(self, register_name: str) -> "Readout":
        return self.data_per_register[register_name]


T = TypeVar("T", bound=Union[int, float, complex])
Readout = List[List[T]]
"""Contains the data of a declared "readout" memory region, usually the ``ro`` register.

All data for all shots for a single Readout will have the same datatype T corresponding to the declared Quil data type:

* ``BIT`` | ``OCTET`` | ``INTEGER``: ``int``
* ``REAL``: ``float``
"""


def create_readout(raw_data: List[List[RawData]]) -> Readout:
    if isinstance(raw_data[0][0], list):
        raw_data = [[complex(entry[0], entry[1]) for entry in shot] for shot in raw_data]

    return raw_data



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##


import logging
import time
import json

from typing import TYPE_CHECKING

from azure.quantum._client.models import JobDetails
from azure.quantum.job.job_failed_with_results_error import JobFailedWithResultsError
from azure.quantum.job.base_job import BaseJob, ContentType, DEFAULT_TIMEOUT
from azure.quantum.job.filtered_job import FilteredJob

__all__ = ["Job", "JobDetails"]

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from azure.quantum.workspace import Workspace


_log = logging.getLogger(__name__)


class Job(BaseJob, FilteredJob):
    """Azure Quantum Job that is submitted to a given Workspace.

    :param workspace: Workspace instance to submit job to
    :type workspace: Workspace
    :param job_details: Job details model,
            contains Job ID, name and other details
    :type job_details: JobDetails
    """

    _default_poll_wait = 0.2

    def __init__(self, workspace: "Workspace", job_details: JobDetails, **kwargs):
        self.results = None
        super().__init__(
            workspace=workspace,
            details=job_details,
            **kwargs
        )

    def submit(self):
        """Submit a job to Azure Quantum."""
        _log.debug(f"Submitting job with ID {self.id}")
        job = self.workspace.submit_job(self)
        self.details = job.details

    def refresh(self):
        """Refreshes the Job's details by querying the workspace."""
        self.details = self.workspace.get_job(self.id).details

    def has_completed(self) -> bool:
        """Check if the job has completed."""
        return (
            self.details.status == "Succeeded"
            or self.details.status == "Failed"
            or self.details.status == "Cancelled"
        )

    def wait_until_completed(
        self,
        max_poll_wait_secs=30,
        timeout_secs=None,
        print_progress=True
    ) -> None:
        """Keeps refreshing the Job's details
        until it reaches a finished status.

        :param max_poll_wait_secs: Maximum poll wait time, defaults to 30
        :type max_poll_wait_secs: int
        :param timeout_secs: Timeout in seconds, defaults to None
        :type timeout_secs: int
        :param print_progress: Print "." to stdout to display progress
        :type print_progress: bool
        :raises TimeoutError: If the total poll time exceeds timeout, raise
        """
        self.refresh()
        poll_wait = Job._default_poll_wait
        start_time = time.time()
        while not self.has_completed():
            if timeout_secs is not None and (time.time() - start_time) >= timeout_secs:
                raise TimeoutError(f"The wait time has exceeded {timeout_secs} seconds.")

            logger.debug(
                f"Waiting for job {self.id},"
                + f"it is in status '{self.details.status}'"
            )
            if print_progress:
                print(".", end="", flush=True)
            time.sleep(poll_wait)
            self.refresh()
            poll_wait = (
                max_poll_wait_secs
                if poll_wait >= max_poll_wait_secs
                else poll_wait * 1.5
            )

    def get_results(self, timeout_secs: float = DEFAULT_TIMEOUT):
        """Get job results by downloading the results blob from the
        storage container linked via the workspace.

        :param timeout_secs: Timeout in seconds, defaults to 300
        :type timeout_secs: float
        :raises RuntimeError: Raises RuntimeError if job execution failed
        :return: Results dictionary with histogram shots, or raw results if not a json object.
        """
        if self.results is not None:
            return self.results

        if not self.has_completed():
            self.wait_until_completed(timeout_secs=timeout_secs)

        if not self.details.status == "Succeeded":
            if self.details.status == "Failed" and self._allow_failure_results():
                job_blob_properties = self.download_blob_properties(self.details.output_data_uri)
                if job_blob_properties.size > 0:
                    job_failure_data = self.download_data(self.details.output_data_uri)
                    raise JobFailedWithResultsError("An error occurred during job execution.", job_failure_data)

            raise RuntimeError(
                f'{"Cannot retrieve results as job execution failed"}'
                + f"(status: {self.details.status}."
                + f"error: {self.details.error_data})"
            )

        payload = self.download_data(self.details.output_data_uri)
        try:
            payload = payload.decode("utf8")
            results = json.loads(payload)

            if self.details.output_data_format == "microsoft.quantum-results.v1":
                if "Histogram" not in results:
                    raise f"\"Histogram\" array was expected to be in the Job results for \"{self.details.output_data_format}\" output format."
                
                histogram_values = results["Histogram"]

                if len(histogram_values) % 2 == 0:
                    # Re-mapping {'Histogram': ['[0]', 0.50, '[1]', 0.50] } to {'[0]': 0.50, '[1]': 0.50}
                    return {histogram_values[i]: histogram_values[i + 1] for i in range(0, len(histogram_values), 2)}
                else: 
                    raise f"\"Histogram\" array has invalid format. Even number of items is expected."
            
            return results
        except:
            # If errors decoding the data, return the raw payload:
            return payload


    @classmethod
    def _allow_failure_results(cls) -> bool: 
        """
        Allow to download job results even if the Job status is "Failed".

        This method can be overridden in derived classes to alter the default
        behaviour.

        The default is False.
        """
        return False


##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##

"""Defines set of targets for interacting with Azure Quantum"""

from .target import Target
from .solvers import Solver
from .toshiba import (
    SimulatedBifurcationMachine
)
from .ionq import IonQ
from .quantinuum import Quantinuum
from .rigetti import Rigetti
from .pasqal import Pasqal
from .microsoft.elements.dft import MicrosoftElementsDft, MicrosoftElementsDftJob

# Default targets to use when there is no target class
# associated with a given target ID
DEFAULT_TARGETS = {
    "ionq": IonQ,
    "quantinuum": Quantinuum,
    "rigetti": Rigetti,
    "toshiba": Solver,
    "pasqal": Pasqal
}


__all__ = [
    "Target",
    "IonQ",
    "Quantinuum",
    "DEFAULT_TARGETS"
    ]